@@ -22,26 +22,19 @@ use std::{
     time::{Duration, UNIX_EPOCH},
 };
 
-use nix::{
-    dir::Type,
-    errno::Errno,
-    sys::mman::{mmap, MapFlags, ProtFlags},
-    sys::stat::Mode,
-    unistd::{Gid, Uid},
-};
-
 use blown_fuse::{
-    io::{Attrs, Entry, FsInfo, Known},
+    io::{Attrs, Entry, EntryType, FsInfo, Gid, Ino, Known, Mode, Stat, Ttl, Uid},
     mount::{mount_sync, Options},
-    ops::{Getattr, Init, Lookup, Readdir, Readlink, Statfs},
+    ops,
     session::{Dispatch, Start},
-    Done, FuseResult, Ino, Op, Ttl,
+    Done, Errno, FuseResult, Op,
 };
 
 use bytemuck::{cast_slice, from_bytes, try_from_bytes};
 use bytemuck_derive::{Pod, Zeroable};
 use clap::{App, Arg};
 use futures_util::stream::{self, Stream, StreamExt, TryStreamExt};
+use nix::sys::mman::{mmap, MapFlags, ProtFlags};
 use smallvec::SmallVec;
 use tokio::{self, runtime::Runtime};
 use uuid::Uuid;
@@ -164,7 +157,7 @@ impl Ext2 {
                 }
 
                 let ino = Ino(header.inode as u64);
-                let name = OsStr::from_bytes(&bytes[..header.name_len as usize]).into();
+                let name = OsStr::from_bytes(&bytes[..header.name_len as usize]);
 
                 let inode = Resolved {
                     ino,
@@ -301,7 +294,7 @@ impl Ext2 {
 }
 
 impl Ext2 {
-    fn init<'o>(&self, (_, reply): Op<'o, Init>) -> Done<'o> {
+    fn init<'o>(&self, (_, reply): Op<'o, ops::Init>) -> Done<'o> {
         let label = &self.superblock.s_volume_name;
         let label = &label[..=label.iter().position(|byte| *byte == b'\0').unwrap_or(0)];
         let label = CStr::from_bytes_with_nul(label)
@@ -323,7 +316,7 @@ impl Ext2 {
         reply.ok()
     }
 
-    async fn statfs<'o>(&self, (_, reply): Op<'o, Statfs>) -> Done<'o> {
+    async fn statfs<'o>(&self, (_, reply): Op<'o, ops::Statfs>) -> Done<'o> {
         let total_blocks = self.superblock.s_blocks_count as u64;
         let free_blocks = self.superblock.s_free_blocks_count as u64;
         let available_blocks = free_blocks - self.superblock.s_r_blocks_count as u64;
@@ -343,88 +336,82 @@ impl Ext2 {
         reply.info(&info)
     }
 
-    async fn getattr<'o>(&self, (request, reply): Op<'o, Getattr>) -> Done<'o> {
+    async fn getattr<'o>(&self, (request, reply): Op<'o, ops::Getattr>) -> Done<'o> {
         let ino = request.ino();
-        let (reply, inode) = reply.fallible(self.inode(ino))?;
+        let (reply, inode) = reply.and_then(self.inode(ino))?;
 
-        reply.known(&Resolved { ino, inode })
+        reply.stat(&Resolved { ino, inode })
     }
 
-    async fn lookup<'o>(&self, (request, reply): Op<'o, Lookup>) -> Done<'o> {
+    async fn lookup<'o>(&self, (request, reply): Op<'o, ops::Lookup>) -> Done<'o> {
         let name = request.name();
-        let (reply, parent) = reply.fallible(self.inode(request.ino()))?;
+        let (mut reply, parent) = reply.and_then(self.inode(request.ino()))?;
 
         //TODO: Indexed directories
-        let lookup = async {
-            let stream = self.directory_stream(parent, 0);
-            tokio::pin!(stream);
+        let stream = self.directory_stream(parent, 0);
+        tokio::pin!(stream);
 
-            loop {
-                match stream.try_next().await? {
-                    Some(entry) if entry.name == name => break Ok(Some(entry.inode)),
-                    Some(_) => continue,
-                    None => break Ok(None),
-                }
+        let inode = loop {
+            let (next_reply, entry) = reply.and_then(stream.try_next().await)?;
+            reply = next_reply;
+
+            match entry {
+                Some(entry) if entry.name == name => break Some(entry.inode),
+                Some(_) => continue,
+                None => break None,
             }
         };
 
-        let (reply, result) = reply.interruptible(lookup).await?;
-        let (reply, inode) = reply.fallible(result)?;
-
         if let Some(inode) = inode {
-            reply.found(inode, Ttl::MAX)
+            reply.known(inode, Ttl::MAX)
         } else {
-            reply.not_found(Ttl::MAX)
+            reply.not_found_for(Ttl::MAX)
         }
     }
 
-    async fn readlink<'o>(&self, (request, reply): Op<'o, Readlink>) -> Done<'o> {
+    async fn readlink<'o>(&self, (request, reply): Op<'o, ops::Readlink>) -> Done<'o> {
         let ino = request.ino();
-        let (reply, inode) = reply.fallible(self.inode(ino))?;
+        let (mut reply, inode) = reply.and_then(self.inode(ino))?;
 
         let resolved = Resolved { ino, inode };
-        if resolved.inode_type() != Type::Symlink {
+        if resolved.inode_type() != EntryType::Symlink {
             return reply.invalid_argument();
         }
 
         let size = inode.i_size as usize;
         if size < size_of::<[u32; 15]>() {
-            return reply.target(OsStr::from_bytes(&cast_slice(&inode.i_block)[..size]));
+            return reply.slice(&cast_slice(&inode.i_block)[..size]);
         }
 
-        let segments = async {
-            /* This is unlikely to ever spill, and is guaranteed not to
-             * do so for valid symlinks on any fs where block_size >= 4096.
-             */
-            let mut segments = SmallVec::<[&[u8]; 1]>::new();
-            let (mut size, mut offset) = (size, 0);
+        /* This is unlikely to ever spill, and is guaranteed not to
+         * do so for valid symlinks on any fs where block_size >= 4096.
+         */
+        let mut segments = SmallVec::<[&[u8]; 1]>::new();
+        let (mut size, mut offset) = (size, 0);
 
-            while size > 0 {
-                let segment = self.seek_contiguous(inode, offset)?;
-                let segment = &segment[..segment.len().min(size)];
+        while size > 0 {
+            let (next_reply, segment) = reply.and_then(self.seek_contiguous(inode, offset))?;
+            reply = next_reply;
 
-                segments.push(segment);
+            let segment = &segment[..segment.len().min(size)];
+            segments.push(segment);
 
-                size -= segment.len();
-                offset += segment.len() as u64;
-            }
+            size -= segment.len();
+            offset += segment.len() as u64;
+        }
 
-            Ok(segments)
-        };
-
-        let (reply, segments) = reply.fallible(segments.await)?;
-        reply.gather_target(&segments)
+        reply.gather(&segments)
     }
 
-    async fn readdir<'o>(&self, (request, reply): Op<'o, Readdir>) -> Done<'o> {
-        let (reply, inode) = reply.fallible(self.inode(request.ino()))?;
+    async fn readdir<'o>(&self, (request, reply): Op<'o, ops::Readdir>) -> Done<'o> {
+        let (reply, inode) = reply.and_then(self.inode(request.ino()))?;
         let mut reply = reply.buffered(Vec::new());
 
         let stream = self.directory_stream(inode, request.offset());
         tokio::pin!(stream);
 
         while let Some(entry) = stream.next().await {
-            let (next_reply, entry) = reply.fallible(entry)?;
+            let (next_reply, entry) = reply.and_then(entry)?;
             let (next_reply, ()) = next_reply.entry(entry)?;
             reply = next_reply;
         }
@@ -433,32 +420,32 @@ impl Ext2 {
     }
 }
 
-impl Known for Resolved {
+impl Stat for Resolved {
     fn ino(&self) -> Ino {
         self.ino
     }
 
-    fn inode_type(&self) -> Type {
+    fn inode_type(&self) -> EntryType {
         let inode_type = self.inode.i_mode >> 12;
         match inode_type {
-            0x01 => Type::Fifo,
-            0x02 => Type::CharacterDevice,
-            0x04 => Type::Directory,
-            0x06 => Type::BlockDevice,
-            0x08 => Type::File,
-            0x0A => Type::Symlink,
-            0x0C => Type::Socket,
+            0x01 => EntryType::Fifo,
+            0x02 => EntryType::CharacterDevice,
+            0x04 => EntryType::Directory,
+            0x06 => EntryType::BlockDevice,
+            0x08 => EntryType::File,
+            0x0A => EntryType::Symlink,
+            0x0C => EntryType::Socket,
 
             _ => {
                 log::error!("Inode {} has invalid type {:x}", self.ino, inode_type);
-                Type::File
+                EntryType::File
             }
         }
     }
 
     fn attrs(&self) -> (Attrs, Ttl) {
         let inode = self.inode;
-        let (access, modify, create) = {
+        let (access, modify, change) = {
             let time = |seconds: u32| (UNIX_EPOCH + Duration::from_secs(seconds.into())).into();
             let (atime, mtime, ctime) = (inode.i_atime, inode.i_mtime, inode.i_ctime);
 
@@ -472,12 +459,21 @@ impl Known for Resolved {
                 Gid::from_raw(inode.i_gid.into()),
             )
             .mode(Mode::from_bits_truncate(inode.i_mode.into()))
-            .blocks(inode.i_blocks.into(), 512)
-            .times(access, modify, create)
+            .blocks(inode.i_blocks.into())
+            .block_size(512)
+            .times(access, modify, change)
             .links(inode.i_links_count.into());
 
         (attrs, Ttl::MAX)
     }
+}
+
+impl Known for Resolved {
+    type Inode = Resolved;
+
+    fn inode(&self) -> &Self::Inode {
+        self
+    }
 
     fn unveil(self) {}
 }
